Package: cellkeyperturbation
Title: Cell Key Perturbation
Version: 3.0.0
Authors@R: c(
    person("Iain", "Dove", email = "SMLHelp@ons.gov.uk", 
    role = c("aut", "cph")),
    person("Ahmet", "Aydin", email = "SDC.Queries@ons.gov.uk", 
    role = c("aut", "cre")))
Description: Provides functions to generate frequency tables and apply cell key
    perturbation to protect against statistical disclosure in tabular outputs.
    The implemented methods are described in "Cell Key Perturbation User Guide"
    <https://github.com/ONSdigital/cell-key-perturbation-R/blob/main/documentation/SML_UserDoc_CKP_R.md>.
    Developed at the UK Office for National Statistics.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: bigrquery, DBI, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Imports: data.table
Depends: R (>= 3.5.0)
LazyData: true
VignetteBuilder: knitr
URL: https://github.com/ONSdigital/cell-key-perturbation-R
BugReports: https://github.com/ONSdigital/cell-key-perturbation-R/issues
NeedsCompilation: no
Packaged: 2026-05-21 10:57:32 UTC; aydina
Author: Iain Dove [aut, cph],
  Ahmet Aydin [aut, cre]
Maintainer: Ahmet Aydin <SDC.Queries@ons.gov.uk>
Repository: CRAN
Date/Publication: 2026-05-28 11:40:06 UTC
Built: R 4.5.2; ; 2026-05-28 12:30:08 UTC; unix

Package: cellkeyperturbation
Title: Cell Key Perturbation
Version: 3.0.0
Authors@R: c(
    person("Iain", "Dove", email = "SMLHelp@ons.gov.uk", 
    role = c("aut", "cph")),
    person("Ahmet", "Aydin", email = "SDC.Queries@ons.gov.uk", 
    role = c("aut", "cre")))
Description: Provides functions to generate frequency tables and apply cell key
    perturbation to protect against statistical disclosure in tabular outputs.
    The implemented methods are described in "Cell Key Perturbation User Guide"
    <https://github.com/ONSdigital/cell-key-perturbation-R/blob/main/documentation/SML_UserDoc_CKP_R.md>.
    Developed at the UK Office for National Statistics.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: bigrquery, DBI, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Imports: data.table
Depends: R (>= 3.5.0)
LazyData: true
VignetteBuilder: knitr
URL: https://github.com/ONSdigital/cell-key-perturbation-R
BugReports: https://github.com/ONSdigital/cell-key-perturbation-R/issues
NeedsCompilation: no
Packaged: 2026-05-21 10:57:32 UTC; aydina
Author: Iain Dove [aut, cph],
  Ahmet Aydin [aut, cre]
Maintainer: Ahmet Aydin <SDC.Queries@ons.gov.uk>
Repository: CRAN
Date/Publication: 2026-05-28 11:40:06 UTC
Built: R 4.5.2; ; 2026-05-28 16:13:52 UTC; unix

# Generated by roxygen2: do not edit by hand

export(create_perturbed_table)
export(create_perturbed_table_bigquery)
export(generate_ptable_10_5_rule)
export(generate_random_rkey)
export(generate_test_data)
export(validate_inputs)
import(data.table)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Cell Key Perturbation

<!-- badges: start -->

[![R-CMD-check](https://github.com/ONSdigital/cell-key-perturbation-R/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ONSdigital/cell-key-perturbation-R/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

This method creates a frequency table which has had cell key
perturbation applied to the counts to protect against disclosure.

Cell key perturbation adds small amounts of noise to frequency tables.
Noise is added to change the counts that appear in the frequency table
by small amounts, for example a 14 is changed to a 15. This noise
introduces uncertainty in the counts and makes it harder to identify
individuals, especially when taking the ‘difference’ between two similar
tables. It protects against the risk of disclosure by differencing since
it cannot be determined whether a difference between two similar tables
represents a real person, or is caused by the perturbation.

Cell Key Perturbation is consistent and repeatable, so the same cells
are always perturbed in the same way.

It is expected that users will tabulate 1 to 4 variables for a
particular geography level - for example, tabulate age by sex at local
authority level. 

### BigQuery

The BigQuery version allows users to perform perturbation without
reading raw data into local memory. The package creates the frequency
table and runs perturbation with an SQL query. Then, it converts the
final perturbed table into a `data.table` as an output.

This will allow users to run the method on large datasets without
breaking the memory limits.

### Terminology

- ***Microdata*** - Data at the level of individual respondents
- ***Record key*** - A random number assigned to each record
- ***Cell value*** - The number of records or frequency for a cell
- ***Cell key*** - The sum of record keys for a given cell
- ***pvalue*** - Perturbation value. The value of noise added to cells,
  e.g. +1, -1
- ***pcv*** - Perturbation cell value. This is an amended cell value
  needed to merge on the ptable
- ***ptable*** - Perturbation table. The look-up file containing the
  pvalues, this determines which cells get perturbed and by how much.

# User Instructions

## Installing the method

This method requires R version 3.5 or higher and uses the `data.table`
package.

You can install the released version of `cellkeyperturbation` from CRAN:

``` r
install.packages("cellkeyperturbation")
```

In your code you can load the cell key perturbation package using:

``` r
library(cellkeyperturbation)
```

## Using the method

You can call the main functions for cell key perturbation with the
following parameters:

``` r
# for data.table
create_perturbed_table(data, ptable, geog, tab_vars, record_key, use_existing_ons_id, threshold)

# for BigQuery
create_perturbed_table_bigquery(con, data, ptable, geog, tab_vars, record_key, use_existing_ons_id, threshold)
```

Parameters specific for BigQuery version:

- **`con`** - (DBIConnection) - An active BigQuery connection created
  with `DBI::dbConnect()`
- **`data`** - (Microdata) - a `character` for the full name of
  micro-level `data` in BigQuery in “\<PROJECT\>.\<DATASET\>.\<TABLE\>”
  format.
- **`ptable`** - (Perturbation table) - a `character` for the full name
  of `ptable` in BigQuery in “\<PROJECT\>.\<DATASET\>.\<TABLE\>” format.

Parameters specific for data.table version:

- **`data`** - (Microdata) - a `data.table` containing the micro-level
  `data` to be tabulated and perturbed.
- **`ptable`** - (Perturbation table) - a `data.table` containing the
  `ptable` file which determines when perturbation is applied.

Common parameters for both versions:

- **`geog`** - (Geography) - a character vector giving the column name
  in `data` that contains the desired geography level you wish to
  tabulate at, e.g. `c("Local_Authority", "Ward")`. This can be the
  empty vector, `geog = c()`, if no geography level is required.
- **`tab_vars`** - (Variables to tabulate) - a character vector giving
  the column names in `data` of the variables to be tabulated
  e.g. `c("Age","Health","Occupation")`. This can also be the empty
  vector, `tab_vars = c()`. However, at least one of `tab_vars` or
  `geog` must be populated. If both are left blank an error message will
  be returned.
- **`record_key`** - a character containing the column name in `data`
  giving the record keys required for perturbation. If `ons_id` is
  available as a column in `data` and `use_existing_ons_id = TRUE`, set
  `record_key = NULL`, as record keys will be generated from `ons_id`.
- **`use_existing_ons_id`** - `TRUE` or `FALSE`, with a default of
  `TRUE`. If `ons_id` is available as a column in `data`, then record
  keys will be derived from `ons_id` by default.
- **`threshold`** - the value below which a count is suppressed (default
  10).

## Worked Example with Synthetic Data (`data.table`)

This is an example showing how to create a perturbed table from
synthetic test data provided in the package (`micro` and `ptable_10_5`).
You can access and view these data tables after loading the package.

``` r
library(cellkeyperturbation)
View(micro)
View(ptable_10_5)
```

You can also generate different sample data or generate random record
keys for testing purposes for your own test data with the following
code:

``` r
data = generate_test_data(size = 1000, rkey_range = 255, seed = 123)
ptable = generate_ptable_10_5_rule(ckey_range = 255)

library(data.table)
data <- fread("input_microdata.csv")
data = generate_random_rkey(data, rkey_range = 255, seed = 123)
```

- **`micro`**: A sample `data.table` containing randomly generated
  microdata and record keys.

Example rows of a microdata table are shown below:

| record_key | var1 | var5 | var8 |
|:-----------|:-----|:-----|:-----|
| 84         | 2    | 9    | D    |
| 108        | 1    | 9    | C    |
| 212        | 1    | 1    | D    |
| 212        | 2    | 2    | A    |
| 86         | 2    | 4    | A    |

- **`ptable_10_5`**: A sample perturbation table (`data.table`) that
  defines the cell key perturbation rules. This specific table applies
  the ’10 to 5 rule’, which means a suppression threshold of *10* and
  rounding to the nearest *5*. In other words, this ptable will remove
  all cells under *10*, and round all others to the nearest *5*.

Example rows of a ptable are shown below:

| pcv | ckey | pvalue |
|:----|:-----|:-------|
| 1   | 0    | -1     |
| 1   | 1    | -1     |
| 1   | 2    | -1     |
| …   | …    | …      |
| 750 | 255  | 0      |

Use the following code to generate the perturbed table using the sample
microdata and perturbation table provided:

``` r
perturbed_table <- create_perturbed_table(
  data       = micro,
  ptable     = ptable_10_5,
  geog       = c("var1"),
  tab_vars   = c("var5","var8"),
  record_key = "record_key",
  threshold  = 10
)
```

## Interpreting the Output

The output from the code is a `data.table` containing a frequency table
with the counts having been affected by perturbation, as specified in
the ptable.

For most ptables, the most obvious effect will be that all counts lower
than the threshold of 10 will have been removed. Suppressing counts
below the threshold is a condition that need to be met when exporting
data from IDS (Integrated Data Service) and many other secure
environments such as SRS (Secure Research Service).

The perturbation code will treat categories for missing data in the same
way as it treats other categories. If you would like to exclude missing
data from your outputs, you will need to remove the missing data
categories either before or after applying the perturbation.

The table will be in the following format:

| var1 | var5 | var8 | pre_sdc_count | ckey | pcv | pvalue | count |
|:-----|:-----|:-----|:--------------|:-----|:----|:-------|:------|
| 1    | 1    | A    | 10            | 173  | 10  | 0      | 10    |
| 1    | 1    | B    | 10            | 88   | 10  | 0      | 10    |
| 1    | 1    | C    | 7             | 180  | 7   | -7     | nan   |
| 1    | 1    | D    | 14            | 66   | 14  | 1      | 15    |
| 1    | 2    | A    | 11            | 190  | 11  | -1     | 10    |
| …    | …    | …    | …             | …    | …   | …      | …     |

The table contains the variables used to summarise the data (in this
example `var1`, `var5` & `var8`), and five other columns:

- `ckey` is the sum of record keys for each combination of variables.
- `pcv` is the perturbation cell value, the pre-perturbation count
  modulo 750.
- `pre_sdc_count` is the pre-perturbation count.
- `pvalue` is the perturbation applied to the original count, most
  commonly it will be 0. This is obtained from the ptable using a join
  on `ckey` and `pcv`.
- `count` is the post-perturbation count, the values to be output. It
  will be set to `NA` if the value is suppressed for being below the
  threshold.

The columns you are most likely interested in are the variables, which
are the categories you’ve summarised by, plus the `count` column.

**WARNING! - The `ckey`, `pcv`, `pre_sdc_count` and `pvalue` columns
should be dropped before the contingency table is published. Otherwise,
the perturbation can be unpicked and the output will be disclosive.**

## Appendix - Help Pages

The package includes further help pages like Introduction to Cell Key
Perturbation vignette and documentation for each function. You can
access these pages by selecting the `cellkeyperturbation` package name
in the packages tab of RStudio or using:

``` r
help(package=cellkeyperturbation)
```

#' Create a frequency table with cell key perturbation applied
#'
#' `create_perturbed_table()` creates a frequency table which has had
#'  cell key perturbation applied to the counts.
#'  A p-table file needs to be supplied which determines which cells are
#'  perturbed.
#'  The data needs to contain a 'record key' variable which along with the
#'  ptable allows the process to be repeatable and consistent.
#'
#' @param data A `data.table` containing the data to be tabulated and perturbed
#'
#'  The data should contain one row per statistical unit (person, household,
#'  business or other) and one column per variable (age, sex, health status)
#' @param ptable A `data.table` containing the `ptable` file which determines
#' when perturbation is applied.
#' @param geog A `character vector` giving the column name in `data` that
#' contains the desired geography level for the frequency table. This can be an
#' empty vector, `c()`, if no geography level is required.
#' @param tab_vars A `character vector` giving the column names in `data` of the
#' variables to be tabulated.
#' This can be an empty vector, `c()`, provided a geography level is supplied.
#' @param record_key A `character` containing the column name in `data` giving
#' the record keys required for perturbation.
#' If data contains "ons_id" and use_existing_ons_id = TRUE,
#' set (record_key = NULL), as record key will be generated from "ons_id".
#' @param use_existing_ons_id A `logical` on whether to create record keys
#' from ons_id, if ons_id exists in data.
#' It will be irrelevant if microdata does not contain ons_id.
#' Default is TRUE.
#' @param threshold An `integer` specifying the value below which counts are
#' suppressed, with a default value of 10.
#'
#' @return Returns a `data.table` giving a frequency table which has had
#' cell key perturbation applied according to the ptable supplied.
#'
#' @import data.table
#'
#' @examples
#' if (requireNamespace("data.table", quietly = TRUE)) {
#'   data.table::setDTthreads(1)
#' }
#'
#' geog <- "var1"
#' tab_vars <- c("var5","var8")
#' record_key <- "record_key"
#' perturbed_table <- create_perturbed_table(micro,
#'                                           ptable_10_5,
#'                                           geog,
#'                                           tab_vars,
#'                                           record_key)
#'
#' # Alternatively
#' perturbed_table <- create_perturbed_table(data = micro,
#'                                           ptable = ptable_10_5,
#'                                           geog = c(),
#'                                           tab_vars = c("var1","var5","var8"),
#'                                           record_key = "record_key",
#'                                           threshold = 10)
#'
#' @export
create_perturbed_table <- function(
    data,
    ptable,
    geog,
    tab_vars,
    record_key,
    use_existing_ons_id = TRUE,
    threshold = 10)
{
  # Generate record keys from "ons_id" if exists
  if (use_existing_ons_id && ("ons_id" %in% colnames(data))) {
    message('NOTE: "ons_id" column is available in data!',
            '\nGenerating record keys from "ons_id"!')

    data <- generate_record_key_from_ons_id(data,
                                            record_key_col = "ons_record_key")
    record_key <- "ons_record_key"
  }

  # Step 0: Validate Inputs
  validate_inputs(data, ptable, geog, tab_vars, record_key, threshold)

  # ============================================================================
  # Rename record_key in case 'record_key' is a column name in the data
  record_key_arg <- record_key

  # Bind variables locally to function to prevent
  # 'No visible binding for global variable' during build check
  pre_sdc_count <- pcv <- count <- pvalue <- ckey <- NULL

  # Step 1: Create frequency table

  #drop unnecessary columns
  data <- data[,c(geog,tab_vars,record_key_arg),with=FALSE]

  #convert every column to factor, except record_key
  cols <- colnames(data)[!(colnames(data) %in% record_key_arg)]
  data[,(cols):=lapply(.SD,as.factor),.SDcols=cols]

  #tabulate - using 'table' function to get zero cells
  aggregated_table <- as.data.table(
    table(
      data[, c(geog, tab_vars), with=FALSE], useNA = "ifany"
    )
  )
  colnames(aggregated_table)[colnames(aggregated_table) == "N"] <-
    "pre_sdc_count"

  # If only 1 variable specified, column is named V1.
  # Rename to original column name to prevent later merge failing.
  if (length(cols) == 1) {
    colnames(aggregated_table)[colnames(aggregated_table) == "V1"] <- cols
  }

  # Step 2: Calculate sum of the record keys and apply modulo to obtain cell key
  max_ckey <- max(ptable$ckey)
  cellkeys <- setDT(data)[,list(ckey = sum(get(record_key_arg))%%(max_ckey+1)),
                          keyby = c(geog,tab_vars)]
  aggregated_table <- merge(aggregated_table, cellkeys, by = c(geog, tab_vars),
                            all.x=TRUE)

  # Step 3: Create pcv by ensuring the rows of ptable 501-750 are reused
  #         for cell values above 750
  aggregated_table$pcv <-
    as.integer(((aggregated_table$pre_sdc_count-1)%%250)+501)
  setDT(aggregated_table)[pre_sdc_count<=750, pcv:=pre_sdc_count,]

  # Step 4: Merge on ptable to get perturbation value for each cell
  aggregated_table <- merge(aggregated_table, ptable, by = c("ckey","pcv"),
                            sort=FALSE, all.x=TRUE)
  setcolorder(aggregated_table, c(cols,"pre_sdc_count","ckey","pcv","pvalue"))

  # Step 5: Apply the perturbation and suppress counts less than the threshold
  setDT(aggregated_table)[,count := pre_sdc_count + pvalue,]

  #replacing NAs in cellkey for zero cells
  aggregated_table[pre_sdc_count==0, ckey:=0]

  #setting pvalue to be zero for zero cells
  aggregated_table[pre_sdc_count==0, pvalue:=0]

  #setting count to be zero for zero cells
  aggregated_table[pre_sdc_count==0, count:=0]

  #setting count to be missing if less than threshold
  aggregated_table[count<threshold, count:=NaN]

  cols <- c("pre_sdc_count","ckey","pcv","pvalue","count")
  aggregated_table[, (cols) := lapply(.SD, as.integer), .SDcols = cols]

  check_for_na(DT = aggregated_table, cols = c(geog, tab_vars))

  return(aggregated_table)
}

# =============================================================================
# -------------------------- TESTING TEMPLATE ---------------------------------
# =============================================================================
# Test
# - 1. type validation on input data & ptable
# - 2. variables have been supplied for tabulation
# - 3. record_key has been specified
# - 4. geog, tab_vars & record_key specified are contained within data
# - 5. ptable has correct format
# - 6. record key range in data and cell key range in ptable match
# - 7. function works when only 1 grouping variable specified
# - 8. validation for missing record keys
# - 9. perturbed table content matches expected output for specific example
# - 10. Check threshold is applied
#
# TEST DATA:
# perturbed_table_var1_var5_var8.rds - expected output for perturbation on
#                                      micro data using ptable_10_5,
#                                      geog=var1 & tab_vars=var5,var8
# =============================================================================

# -----------------------------------------------------------------------------
# TESTS:1. Check type validation on input data & ptable
# -----------------------------------------------------------------------------

test_that("error raised if input data NOT a data.table", {
  expect_error(create_perturbed_table(data = "not_a_data_table",
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = "record_key",
                                      ptable = ptable_10_5),
        "Specified value for data must be a data.table", fixed = TRUE)
})

test_that("error raised if ptable supplied NOT a data.table", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = "record_key",
                                      ptable = "not_a_data_table"),
        "Specified value for ptable must be a data.table", fixed = TRUE)
})

# -----------------------------------------------------------------------------
# TESTS:2. Check error raised if geog & tab_vars are both NULL or empty vectors
# -----------------------------------------------------------------------------

test_that("error raised if geog & tab_vars are both NULL.", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = NULL,
                                      tab_vars = NULL,
                                      record_key = "record_key",
                                      ptable = ptable_10_5),
  "No variables for tabulation. Specify value for geog or tab_vars.",
  fixed = TRUE)
})

test_that("error raised if both geog & tab_vars are empty.", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c(),
                                      tab_vars = c(),
                                      record_key = "record_key",
                                      ptable = ptable_10_5),
  "No variables for tabulation. Specify value for geog or tab_vars.",
  fixed = TRUE)
})

# -----------------------------------------------------------------------------
# TESTS:3 Check error raised if record_key not specified as character
# -----------------------------------------------------------------------------

test_that("error raised if record_key not specified correctly", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = 3,
                                      ptable = ptable_10_5),
               "Expected 'record_key' to be character or NULL, but got 'double'!",
               fixed = TRUE)
})
test_that("error raised if record_key not specified correctly", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = c("var2","var3"),
                                      ptable = ptable_10_5),
               "Specified value for record_key must be a single vector.",
               fixed = TRUE)
})

# -----------------------------------------------------------------------------
# TESTS:4 Check errors raised if geog, tab_vars & record_key not within data
# -----------------------------------------------------------------------------

test_that("error raised if geog not a column within data", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("not_col_in_data"),
                                      tab_vars = c("var5","var8"),
                                      record_key = "record_key",
                                      ptable = ptable_10_5),
        "Specified value for geog must be a column name in data.",
        fixed = TRUE)
})

test_that("error raised if tab_vars not columns within data", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("var1"),
                                      tab_vars = c("not_col_in_data","var8"),
                                      record_key = "record_key",
                                      ptable = ptable_10_5),
        "Specified values for tab_vars must be column names in data.",
        fixed = TRUE)
})

test_that("error raised if record_key not a column within data", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = "not_record_key",
                                      ptable = ptable_10_5),
          "Specified value for record_key must be a column name in data.",
          fixed = TRUE)
})

# -----------------------------------------------------------------------------
# TESTS: 5. Check errors raised if ptable has incorrect format
# -----------------------------------------------------------------------------

dodgy_ptable <- data.table(
  pcv = 1:5,
  ckey = 6:10,
  zvalue = 11:15
)

test_that("error raised if ptable does not contain required column names", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = "record_key",
                                      ptable = dodgy_ptable),
        "Supplied ptable must contain columns named 'pcv','ckey' and 'pvalue'",
        fixed = TRUE)
})

# -----------------------------------------------------------------------------
# TESTS: 6. Check warning given if max record key in data and max cell key
#           in ptable are different
# -----------------------------------------------------------------------------

# Create dataset with record_key up to 254 (ptable_10_5 has max value of
# cell key as 255)
dodgy_data <- data.table(
  record_key = 0:254,
  var1 = 1:255,
  var2 = 11:265
)

test_that("warning if max record key and max cell key are different", {
  expect_warning(create_perturbed_table(data = dodgy_data,
                                      geog = c("var1"),
                                      tab_vars = c("var2"),
                                      record_key = "record_key",
                                      ptable = ptable_10_5))
})

# -----------------------------------------------------------------------------
# TESTS: 7. Check function works when only 1 grouping variable specified
# -----------------------------------------------------------------------------

test_that("create_perturbed_table works when only 1 grouping variable", {

  result <- create_perturbed_table(data = micro,
                                   geog = NULL,
                                   tab_vars = c("var5"),
                                   record_key = "record_key",
                                   ptable = ptable_10_5,
                                   threshold=10)
  expect_equal(nrow(result),10) # var5 values 1-10


  result <- create_perturbed_table(data = micro,
                                   geog = c("var1"),
                                   tab_vars = NULL,
                                   record_key = "record_key",
                                   ptable = ptable_10_5,
                                   threshold=10)
  expect_equal(nrow(result),5) # var1 values 1-5

})

# -----------------------------------------------------------------------------
# TESTS: 8. Check Error/Warning if record keys missing
# -----------------------------------------------------------------------------

micro_test = copy(micro)            # take explicit copy to ensure micro unchanged
micro_test[1000,"record_key":=255]  # make sure max record key not deleted

message_string <- "Less than 50% of records have a record key.
    Cell key perturbation will be much less effective with fewer record keys,
    so this code requires at least 50% of records to have a record key."

micro_test[1:501,"record_key":=NaN]  # Set more than half record keys to missing
test_that("error raised if more then 50% record keys missing", {
  expect_error(create_perturbed_table(data = micro_test,
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = "record_key",
                                      ptable = ptable_10_5),
               message_string,
               fixed = TRUE)
})

micro_test[,"record_key":=NaN]  # Set all record keys to missing
test_that("error raised if more then 50% record keys missing", {
    expect_error(create_perturbed_table(data = micro_test,
                                        geog = c("var1"),
                                        tab_vars = c("var5","var8"),
                                        record_key = "record_key",
                                        ptable = ptable_10_5),
                 message_string,
                 fixed = TRUE)
})

micro_test = copy(micro)
micro_test[1:5,"record_key":=NaN]  # Set 5 record keys to missing
test_that("warning if less than 50% record keys missing" , {
  expect_warning(create_perturbed_table(data = micro_test,
                                        geog = c("var1"),
                                        tab_vars = c("var2"),
                                        record_key = "record_key",
                                        ptable = ptable_10_5))
})

rm(micro_test)


# -----------------------------------------------------------------------------
# TESTS: 9. Check perturbed table content matches expected output
# -----------------------------------------------------------------------------

# Create perturbed table for micro data using geog=var1 & tab_vars=var5,var8
# which includes zero count combinations but should return a table with all
# combinations i.e. 200 rows
test_that("Perturbed table includes zero count combinations", {

  result <- create_perturbed_table(data = micro,
                                   geog = c("var1"),
                                   tab_vars = c("var5","var8"),
                                   record_key = "record_key",
                                   ptable = ptable_10_5,
                                   threshold=10)

  expect_equal(nrow(result),200)
})


# Create perturbed table for micro data with geog=var1 & tab_vars=var5,var8
# and check columns, including counts, match those from python
# implementation of ckp
test_that("Perturbed table has expected counts", {

  result <- create_perturbed_table(data = micro,
                                   geog = c("var1"),
                                   tab_vars = c("var5","var8"),
                                   record_key = "record_key",
                                   ptable = ptable_10_5,
                                   threshold=10)


  # Load expected_result from rda file, and sort data for comparison
  load(file = "perturbed_table_var1_var5_var8.rda")
  grouping_cols <- c("var1","var5","var8")
  expected_result[,(grouping_cols):=lapply(.SD,as.character),
                  .SDcols=grouping_cols]
  setorderv(expected_result,grouping_cols)
  setorderv(result,grouping_cols)

  # Compare calculated columns, including counts
  expect_identical(result$ckey,expected_result$ckey)
  expect_identical(result$pcv,expected_result$pcv)
  expect_identical(result$pre_sdc_count,expected_result$pre_sdc_count)
  expect_identical(result$pvalue,expected_result$pvalue)
  expect_identical(result$count,expected_result$count)
})

# -----------------------------------------------------------------------------
# TESTS: 10. Check threshold is applied
# -----------------------------------------------------------------------------

# Create perturbed table for micro data using geog=var1 & tab_vars=var5,var8
# without specifying value for threshold, check default is applied
test_that("Default threshold is applied", {

  result <- create_perturbed_table(data = micro,
                                   geog = c("var1"),
                                   tab_vars = c("var5","var8"),
                                   record_key = "record_key",
                                   ptable = ptable_10_5)

  expect_true(sum(is.na(result$count))>0)
})

# Create perturbed table for micro data using geog=var1 & tab_vars=var5,var8
# specifying non-integer for threshold, check default is applied
message_string <- "Specified value for threshold must be an integer"

test_that("Error if threshold not integer", {
  expect_error(create_perturbed_table(data = micro,
                                      geog = c("var1"),
                                      tab_vars = c("var5","var8"),
                                      record_key = "record_key",
                                      ptable = ptable_10_5,
                                      threshold = 3.5),
             message_string,
             fixed = TRUE)
})

#' Create a perturbed frequency table in BigQuery and return it as a data frame
#'
#' This function runs the perturbation method fully in BigQuery (via SQL) and
#' only downloads the result, which allows handling large datasets efficiently.
#'
#' @details
#' Function workflow:
#'   1) Generate BigQuery SQL query to run perturbation on BigQuery
#'   2) If return_query = TRUE, return the query text and exit:
#'      otherwise, execute the rest
#'   3) Validate inputs using BigQuery
#'   4) Run perturbation using BigQuery
#'   5) Convert perturbed table to data.table and sort
#'
#' The query build by this function does the following when executed:
#'  * Computes counts and cell keys for each unique combination of geographic
#' and tabulation variables.
#'  * Includes zero-count cells by generating the full cartesian product of
#' variable combinations.
#'  * Calculates pcv by ensuring the rows of ptable 501-750 are reused for
#' cell values above 750.
#'  * Applies perturbation values from a perturbation table based on cell keys
#' and pseudo cell values (pcv).
#'  * Suppresses cells below a specified threshold by setting their perturbed
#' count to NULL.
#'
#' @param con --`DBIConnection`.
#'   An active BigQuery connection created with `DBI::dbConnect()`
#' @param data --`character`.
#'   BigQuery table name for microdata in full format:
#'   `"<PROJECT>.<DATASET>.<TABLE>"`.
#'   One row per statistical unit (person, household, business, etc.),
#'   and one column per variable (e.g. age, sex, health status)
#' @param ptable --`character`.
#'   BigQuery table name for the p-table in full format:
#'   `"<PROJECT>.<DATASET>.<TABLE>"`.
#' @param geog --`character vector`.
#'   Column name containing the desired geography level for the frequency table.
#'   e.g., `c("Region")` or `c("LocalAuthority")`.
#'   Use `c()` if no geography breakdown required.
#' @param tab_vars --`character vector`.
#'   Column names to tabulate, e.g., `c("Age", "Health", "Occupation")`.
#' @param record_key --`character`.
#'   Column name with record keys required for perturbation,
#'   e.g., `"Record_Key"`.
#'   If data contains "ons_id" and use_existing_ons_id = TRUE,
#'   set (record_key = NULL), as record key will be generated from "ons_id".
#' @param use_existing_ons_id -- `logical`
#'   Whether to create record keys from ons_id, if ons_id exists in data.
#'   It will be irrelevant if microdata does not contain ons_id.
#'   Default is TRUE.
#' @param threshold --`integer`.
#'   Suppression threshold; perturbed counts below this value are suppressed.
#'   Default 10.
#' @param return_query --`logical`.
#'   If `TRUE`, returns the generated SQL query without executing it.
#'   Default `FALSE`.
#'
#' @return
#' - When `return_query = FALSE`: a `data.table` containing the perturbed
#'   frequency table, sorted by `geog` and `tab_vars`.
#' - When `return_query = TRUE`: a character string containing the query.
#'
#' @import data.table
#'
#' @export
#'
#' @examples
#' # --- Return query text without executing it ---
#' query <- create_perturbed_table_bigquery(
#'   con        = NULL,
#'   data       = "my-gcp-project.survey.microdata",
#'   ptable     = "my-gcp-project.sdc.ptable",
#'   geog       = c("Region"),
#'   tab_vars   = c("AgeGroup", "HealthStatus", "Occupation"),
#'   record_key = "Record_Key",
#'   threshold  = 10,
#'   return_query = TRUE
#' )
#' cat(query)
#'
create_perturbed_table_bigquery <- function(
    con,
    data,
    ptable,
    geog,
    tab_vars,
    record_key,
    use_existing_ons_id = TRUE,
    threshold = 10,
    return_query = FALSE)
{
  query <- build_perturbation_bigquery(
    data       = data,
    ptable     = ptable,
    geog       = geog,
    tab_vars   = tab_vars,
    record_key = record_key,
    threshold  = threshold
  )

  if (return_query) {
    return(query)
  }

  if (!requireNamespace("DBI", quietly = TRUE)) {
    stop(
      'Package "DBI" must be installed to validate inputs and execute query.',
      call. = FALSE
    )
  }
  if (!inherits(con, "DBIConnection")) {
    stop(
      "`con` must be a DBIConnection when `return_sql = FALSE`.",
      call. = FALSE)
  }

  # Update query to generate record keys from "ons_id" if exists
  data_schema <- DBI::dbGetQuery(con, sprintf("SELECT * FROM `%s` LIMIT 0",
                                              data))

  if (use_existing_ons_id && ("ons_id" %in% colnames(data_schema))) {
    message('NOTE: "ons_id" column is available in data!',
            '\nUpdating query to generate record keys from "ons_id"!')

    query <- gsub(paste0("SAFE_CAST(", record_key, " AS INT64)"),
                  "MOD(SAFE_CAST(ons_id AS INT64), 4096)",
                  query, fixed = TRUE)
  }

  validate_inputs_bigquery(
    con        = con,
    data       = data,
    ptable     = ptable,
    geog       = geog,
    tab_vars   = tab_vars,
    record_key = record_key,
    use_existing_ons_id = use_existing_ons_id,
    threshold  = threshold
  )

  perturbed_table <- DBI::dbGetQuery(con, query)

  perturbed_table <- as.data.table(perturbed_table)

  cols <- c("pre_sdc_count","ckey","pcv","pvalue","count")
  perturbed_table[, (cols) := lapply(.SD, as.integer), .SDcols = cols]

  sort_cols <- c(geog, tab_vars)
  setorderv(perturbed_table, sort_cols)

  check_for_na(DT = perturbed_table, cols = c(geog, tab_vars))

  return(perturbed_table)
}


##==============================================================================
build_perturbation_bigquery <- function(
    data,
    ptable,
    geog,
    tab_vars,
    record_key,
    threshold)
{
  all_vars <- c(geog, tab_vars)

  # Build common strings
  all_vars_str   <- paste(all_vars, collapse = ", ")
  dim_ctes       <- paste0("dim_", all_vars, " AS (SELECT DISTINCT ",
                           all_vars, " FROM distinct_vars)")
  dim_ctes_str   <- paste(dim_ctes, collapse = ",\n\t")
  cross_join     <- paste(paste0("dim_", all_vars),
                          collapse = "\n\tCROSS JOIN ")
  join_conditions <- paste(paste0("g.", all_vars,
                                  " IS NOT DISTINCT FROM b.", all_vars),
                           collapse = "\n\t\tAND ")
  select_columns  <- paste(paste0("g.", all_vars), collapse = ", ")

  # Compose query
  query <- paste0("
-- Step 1: Create dimension tables
    WITH
        distinct_vars AS (
            SELECT DISTINCT ", all_vars_str, "
            FROM `", data, "`
        ),
        ", dim_ctes_str, ",

-- Step 2: Create full grid of all combinations
    full_grid AS (
        SELECT *
        FROM ", cross_join, "
    ),

-- Step 3: Aggregate actual counts
    base_counts AS (
        SELECT
            ", all_vars_str, ",
            COUNT(*) AS pre_sdc_count,
            SUM(SAFE_CAST(", record_key, " AS INT64)) AS sum_rkey
        FROM `", data, "`
        GROUP BY ", all_vars_str, "
    ),

-- Step 4: Join full grid with actual counts
    full_counts AS (
        SELECT
            ", select_columns, ",
            COALESCE(b.pre_sdc_count, 0) AS pre_sdc_count,
            COALESCE(b.sum_rkey, 0) AS sum_rkey
        FROM full_grid g
        LEFT JOIN base_counts b
            ON ", join_conditions, "
    ),

-- Step 5: Compute cell key modulo
    ckey_mod AS (
        SELECT *,
            MOD(sum_rkey, (SELECT MAX(ckey) + 1
              FROM `", ptable, "`)) AS ckey
        FROM full_counts
    ),

-- Step 6: Calculate pcv
    pcv_calc AS (
        SELECT *,
            CASE
                WHEN pre_sdc_count <= 750 THEN pre_sdc_count
                ELSE MOD((pre_sdc_count - 1), 250) + 501
            END AS pcv
        FROM ckey_mod
    ),

-- Step 7: Join with perturbation table
    joined AS (
        SELECT
            ", all_vars_str, ",
            a.pre_sdc_count,
            a.ckey,
            a.pcv,
            COALESCE(b.pvalue, 0) AS pvalue
        FROM pcv_calc a
        LEFT JOIN `", ptable, "` b
            ON a.pcv = b.pcv AND a.ckey = b.ckey
    ),

-- Step 8: Apply perturbation and suppression
    final_table AS (
        SELECT *,
            pre_sdc_count + pvalue AS raw_count,
            CASE
                WHEN pre_sdc_count + pvalue < ", threshold, " THEN NULL
                ELSE pre_sdc_count + pvalue
            END AS count
        FROM joined
    )

-- Final output
    SELECT
        ", all_vars_str, ",
        pre_sdc_count,
        ckey,
        pcv,
        pvalue,
        count
    FROM final_table;"
  )
  return(query)
}

#' Validate Inputs Before Perturbation
#'
#' @description
#' Validates inputs for a perturbation process.
#'
#' * Validate type of input data & ptable
#' * Validate other input arguments
#'   * Check that at least one variable specified for geog or tab_vars
#'   * Check geog and tab_vars are either character vectors or NULL
#'   * Check specified record_key is character vector or NULL
#'   * Check threshold is an integer and non-negative
#' * Validate microdata and ptable contain required columns
#'   * Check data contain the specified geog, tab_vars & record_key
#'   * Check ptable contains required columns
#' * Validate the range of record keys and cell keys
#' * Validate data has sufficient records with record keys to apply perturbation
#'
#' @import data.table
#'
#' @inheritParams create_perturbed_table
#'
#' @return Invisibly returns TRUE on success. Throws stop or Warning messages
#'  if any validation fails.
#'
#' @export
#' @keywords internal
#'
#' @examples
#' if (requireNamespace("data.table", quietly = TRUE)) {
#'   data.table::setDTthreads(1)
#' }
#' validate_inputs(data = micro,
#'                 ptable = ptable_10_5,
#'                 record_key = "record_key",
#'                 geog = c("var1"),
#'                 tab_vars = c("var5","var8"),
#'                 threshold = 10)
validate_inputs <- function(
    data,
    ptable,
    geog,
    tab_vars,
    record_key,
    threshold)
{
  record_key_arg <- record_key

  check_input_data_types(data, ptable)

  check_input_arguments(geog, tab_vars, record_key_arg, threshold)

  check_data_contain_columns(data, geog, tab_vars, record_key_arg)
  check_ptable_contain_columns(ptable)

  # Check if the range of record keys and cell keys match
  min_ckey <- min(ptable$ckey)
  max_ckey <- max(ptable$ckey)
  min_rkey <- min(data[[record_key_arg]], na.rm = TRUE)
  max_rkey <- max(data[[record_key_arg]], na.rm = TRUE)

  check_key_range(min_ckey, max_ckey, min_rkey, max_rkey)

  # Check data has sufficient % records with record keys to apply perturbation
  rkey_na_count <- sum(is.na(data[,get(record_key_arg)]))
  rkey_percent <- 100*(1 - rkey_na_count/nrow(data))

  check_missing_record_key(rkey_na_count, rkey_percent)

  invisible(TRUE)
}

## =============================================================================
#' Validate Inputs Before Perturbation using BigQuery
#'
#' @description
#' Validates BigQuery inputs for a perturbation process.
#'
#' * Validate input arguments
#'   * Check that at least one variable specified for geog or tab_vars
#'   * Check geog and tab_vars are either character vectors or NULL
#'   * Check specified record_key is character vector or NULL
#'   * Check threshold is an integer and non-negative
#' * Validate microdata and ptable contain required columns
#'   * Check data contain the specified geog, tab_vars & record_key
#'   * Check ptable contains required columns
#' * Validate the range of record keys and cell keys
#' * Validate data has sufficient records with record keys to apply perturbation
#'
#' @import data.table
#'
#' @inheritParams create_perturbed_table_bigquery
#'
#' @return Invisibly returns TRUE on success. Throws stop or Warning messages
#'  if any validation fails.
#'
validate_inputs_bigquery <- function(
    con,
    data,
    ptable,
    geog,
    tab_vars,
    record_key,
    use_existing_ons_id,
    threshold)
{
  if (!requireNamespace("DBI", quietly = TRUE)) {
    stop(
      "Package \"DBI\" must be installed to use this function.",
      call. = FALSE
    )
  }

  data_schema <- DBI::dbGetQuery(con, sprintf("SELECT * FROM `%s` LIMIT 0",
                                              data))
  if (use_existing_ons_id && ("ons_id" %in% colnames(data_schema))) {
    record_key = "NULL"
  }

  # 1) Validate input arguments
  if (!is.character(data) | !is.character(ptable)) {
    stop("'data' and 'ptable' must be character type,
         specifying location of tables in BigQuery database!")
  }
  check_input_arguments(geog, tab_vars, record_key, threshold)


  # 2) Check microdata contains required columns
  required_columns <- c(geog, tab_vars, record_key)
  existing_columns <- colnames(data_schema)

  if (use_existing_ons_id && ("ons_id" %in% colnames(data_schema))) {
    required_columns <- c(geog, tab_vars)
  }

  missing <- setdiff(required_columns, existing_columns)
  if (length(missing) > 0) {
    stop(sprintf("Missing columns in '%s': %s",
                 data, paste(missing, collapse = ", ")),
         call. = FALSE)
  }


  # 3) Check ptable contains required columns: ckey, pcv, pvalue
  ptable_schema_df <- DBI::dbGetQuery(con, sprintf("SELECT * FROM `%s` LIMIT 0",
                                                   ptable))
  ptable_cols <- colnames(ptable_schema_df)

  for (col in c("ckey", "pcv", "pvalue")) {
    if (!(col %in% ptable_cols)) {
      stop(sprintf("Missing column '%s' in perturbation table '%s'.",
                   col, ptable),
           call. = FALSE)
    }
  }


  # 4) Check if the range of record keys and cell keys match
  range_query <- sprintf("
                          WITH
                            data_range AS (
                              SELECT
                                MIN(SAFE_CAST(%s AS INT64)) AS min_rkey,
                                MAX(SAFE_CAST(%s AS INT64)) AS max_rkey
                              FROM `%s`
                            ),
                            ptable_range AS (
                              SELECT
                                MIN(ckey) AS min_ckey,
                                MAX(ckey) AS max_ckey
                              FROM `%s`
                            )
                          SELECT
                            d.min_rkey,
                            d.max_rkey,
                            p.min_ckey,
                            p.max_ckey
                          FROM data_range d, ptable_range p
                         ", record_key, record_key, data, ptable)

  if (use_existing_ons_id && ("ons_id" %in% colnames(data_schema))) {
    range_query <- gsub(paste0("SAFE_CAST(", record_key, " AS INT64)"),
                        "MOD(SAFE_CAST(ons_id AS INT64), 4096)",
                        range_query, fixed = TRUE)
  }

  keys_range <- DBI::dbGetQuery(con, range_query)

  min_rkey <- keys_range[["min_rkey"]][1]
  max_rkey <- keys_range[["max_rkey"]][1]
  min_ckey <- keys_range[["min_ckey"]][1]
  max_ckey <- keys_range[["max_ckey"]][1]

  check_key_range(min_ckey, max_ckey, min_rkey, max_rkey)


  # 5) Check data has sufficient records with record keys to apply perturbation
  records_key_query <- sprintf("
    SELECT
      COUNT(*) AS total_records,
      COUNTIF(SAFE_CAST(%s AS INT64) IS NULL) AS null_record_keys,
      ROUND(100.0 * COUNTIF(SAFE_CAST(%s AS INT64) IS NOT NULL) / COUNT(*), 2) AS percent_with_keys
    FROM `%s`
  ", record_key, record_key, data)

  if (use_existing_ons_id && ("ons_id" %in% colnames(data_schema))) {
    records_key_query <- gsub(paste0("SAFE_CAST(", record_key, " AS INT64)"),
                              "MOD(SAFE_CAST(ons_id AS INT64), 4096)",
                              records_key_query, fixed = TRUE)
  }

  rkey <- DBI::dbGetQuery(con, records_key_query)

  rkey_nan_count <- rkey[["null_record_keys"]][1]
  rkey_percent   <- rkey[["percent_with_keys"]][1]

  check_missing_record_key(rkey_nan_count, rkey_percent)

  message("Validation checks are completed!")

  invisible(TRUE)
}


## =============================================================================
## Helper functions

check_input_data_types <- function(data, ptable)
{
  if (!is.data.table(data)) {
    stop("Specified value for data must be a data.table.")
  }
  if (!is.data.table(ptable)) {
    stop("Specified value for ptable must be a data.table.")
  }
}


check_input_arguments <- function(geog, tab_vars, record_key_arg, threshold)
{
  if (length(geog)==0 & length(tab_vars)==0)  {
    stop("No variables for tabulation. Specify value for geog or tab_vars.")
  }
  if (!is.null(geog) && !is.character(geog)) {
    stop(sprintf(
      "Expected 'geog' to be character or NULL, but got '%s'!",
      typeof(geog)
    ))
  }
  if (!is.null(tab_vars) && !is.character(tab_vars)) {
    stop(sprintf(
      "Expected 'tab_vars' to be character or NULL, but got '%s'!",
      typeof(tab_vars)
    ))
  }
  if (!is.null(record_key_arg) && !is.character(record_key_arg)) {
    stop(sprintf(
      "Expected 'record_key' to be character or NULL, but got '%s'!",
      typeof(record_key_arg)
    ))
  }
  if (length(record_key_arg)>1) {
    stop("Specified value for record_key must be a single vector.")
  }
  if ((!is.numeric(threshold) | !(round(threshold)==threshold))){
    stop("Specified value for threshold must be an integer")
  }
  if (threshold <0){
    warning("Specified value for threshold is negative,
            meaning no threshold will be applied.")
  }
}


check_data_contain_columns <- function(data, geog, tab_vars, record_key_arg)
{
  if (length(geog)>0){
    if (!(geog %in% colnames(data))){
      stop("Specified value for geog must be a column name in data.")
    }
  }
  if (length(tab_vars)>0){
    if (!all(tab_vars %in% colnames(data))){
      stop("Specified values for tab_vars must be column names in data.")
    }
  }
  if (!(record_key_arg %in% colnames(data))){
    stop("Specified value for record_key must be a column name in data.")
  }
}


check_ptable_contain_columns <- function(ptable)
{
  msg <- "Supplied ptable must contain columns named 'pcv','ckey' and 'pvalue'."
  if (!all(c("pcv","ckey","pvalue") %in% colnames(ptable))){
    stop(msg)
  }
}


check_key_range <- function(min_ckey, max_ckey, min_rkey, max_rkey)
{
  if (max_ckey != max_rkey){
    warning(paste("Record key and cell key ranges are different.",
                  "The maximum record key is",max_rkey,
                  ", whereas the maximum cell key is",max_ckey,
                  ". Please check you are using the appropriate ptable for this
                  data."))
  }
  if (min_ckey < 0){
    warning("Negative cell key found in ptable!")
  }
  if (min_rkey < 0){
    warning("Negative record key found in data!")
  }
}


check_missing_record_key <- function(rkey_na_count, rkey_percent)
{
  if (rkey_percent < 50){
    message_string <- "Less than 50% of records have a record key.
    Cell key perturbation will be much less effective with fewer record keys,
    so this code requires at least 50% of records to have a record key."
    stop(message_string)
  }
  else if (rkey_percent < 100){
    warning_string <- paste(
      sprintf("Only %.1f%% of records have record keys.",round(rkey_percent,1)),
      sprintf("\n%s record(s) have missing record keys.", rkey_na_count)
    )
    warning(warning_string)
  }
}

#' Generate ptable (10-5 rule)
#'
#' `generate_ptable_10_5_rule()` generates a sample p-table based on 10-5 rule,
#'  which means a suppression threshold of 10 and rounding to the nearest 5.
#'
#' @param max_pcv  Max value for pcv. Default is 750.
#' @param ckey_range The max range for cell keys. Default is 255.
#'
#' @return A data.table assigning a pvalue to each ckey and pcv combination
#'
#' @examples
#' if (requireNamespace("data.table", quietly = TRUE)) {
#'   data.table::setDTthreads(1)
#' }
#' ptable <- generate_ptable_10_5_rule()
#'
#' @import data.table
#'
#' @export
generate_ptable_10_5_rule <- function(max_pcv = 750, ckey_range = 255) {
  DT <- CJ(pcv = 1:max_pcv, ckey = 0:ckey_range, unique = TRUE)

  pval <- as.integer(calculate_pvalue(DT[["pcv"]]))
  set(DT, j = "pvalue", value = pval)

  setkeyv(DT, c("pcv", "ckey"))
  return(DT)
}

#' Calculate pvalue for each pcv based on 10-5 rule
#'
#' @param pcv Perturbation cell value
#'
#' @return Perturbation value, i.e. noise added to cells
#'
#' @noRd
#'
calculate_pvalue <- function(pcv) {
  ifelse(
    pcv < 10,
    -pcv,
    c(`0` = 0, `1` = -1, `2` = -2, `3` = 2, `4` = 1)[as.character(pcv %% 5)]
  )
}

#' Check perturbed table for missingness in tabulation variables
#'
#' @param DT -- `data.table`
#'  Perturbed frequency table
#' @param cols -- `character vector`
#'  Tabulation variables
#'
#' @returns Warning message if any tabulation variable contain missing values
#'
#' @import data.table
#'
check_for_na <- function(DT, cols) {
  na_cols <- DT[, names(.SD)[sapply(.SD, function(x) any(is.na(x)))],
                .SDcols = cols]

  if (length(na_cols) > 0) {
    warning(sprintf("Missing values detected in the following variable(s): %s
      Consider removing frequencies for missing values from the output table.",
      paste(na_cols, collapse = ", ")),
      call. = FALSE)
  }
}

#' Generate Record Key from ONS ID
#'
#' This function creates a new record key column by taking the modulo 4096 of
#' the `ons_id` column. It converts `ons_id` to numeric, preserving `NA` for
#' non-numeric values, and assigns the result as an integer.
#'
#' @param data A `data.table` containing the `ons_id` column.
#' @param record_key_col A character string specifying the name of the new
#'   record key column to create.
#'
#' @return A `data.table` with the new record key column added.
#'
#' @details
#' - The function checks that `data` is a `data.table`.
#' - Non-numeric values in `ons_id` are converted to `NA`.
#' - The record key is computed as `ons_id %% 4096` and stored as integer.
#'
#' @import data.table
#'
generate_record_key_from_ons_id <- function(data, record_key_col) {

  if (!data.table::is.data.table(data)) {
    stop("data must be a data.table")
  }

  df <- data.table::copy(data)

  df[, (record_key_col) := as.integer(as.numeric(get("ons_id")) %% 4096L)]

  return(df)
}


#' Generate and attach random record keys to microdata
#'
#' @description
#' `generate_random_key()` attaches randomly generated record keys to microdata
#'  tables for testing purposes.
#'
#' @import data.table
#'
#' @param data A data.table or data.frame containing the microdata
#' @param rkey_range The max range for record keys. Default is 255.
#' @param seed A seed for the random number generator
#'
#' @return A data.table with a new integer column `record_key`
#'
#' @export
#'
#' @examples
#' library(data.table)
#' data <- data.table(id = 1:1000)
#' data <- generate_random_rkey(data, rkey_range = 255, seed = 2005)
generate_random_rkey <- function(data, rkey_range = 255, seed = NULL) {

  if (!data.table::is.data.table(data)) {
    stop("Input data must be a 'data.table'!")
  }
  dt <- copy(data)

  if(!is.null(seed)){
    set.seed(seed)
  }

  dt[, ("record_key") := sample(0:rkey_range, .N, replace = TRUE)]

  return(dt)
}

#' Perturbation table
#'
#'  A data set containing the rules to apply cell key perturbation with a
#'  threshold of 10, and rounding to base 5. In other words, counts less than
#'  10 will be removed, and all others will be rounded to the nearest 5.
#'
#' \itemize{
#'   \item pcv. perturbation cell value (1-750)
#'   \item ckey. cell key value (0-255)
#'   \item pvalue. perturbation value to be applied
#' }
#'
#' @docType data
#' @keywords datasets
#' @name ptable_10_5
#' @usage data(ptable_10_5)
#' @format A data.table containing 192000 observations of 3 variables
"ptable_10_5"


#' Example data (micro)
#'
#'  A data set containing randomly generated data to showcase the cell key
#'  perturbation method.
#'
#' \itemize{
#'   \item record_key. record key value (0-255)
#'   \item var1. example variable 1 (1-5)
#'   \item var2. example variable 2 (1,2)
#'   \item var3. example variable 3 (1-4)
#'   \item var4. example variable 4 (1-4)
#'   \item var5. example variable 5 (1-10)
#'   \item var6. example variable 6 (1-5)
#'   \item var7. example variable 7 (1-5)
#'   \item var8. example variable 8 (A-D)
#'   \item var9. example variable 9 (A-H)
#'   \item var10. example variable 10 (1-49)
#' }
#'
#' @docType data
#' @keywords datasets
#' @name micro
#' @usage data(micro)
#' @format A data.table containing 1000 observations of 11 variables
"micro"

#' Generate sample microdata
#'
#' @description
#' `generate_test_data()` creates a sample microdata containing randomly
#'  generated microdata columns and record keys for testing purposes.
#'
#' Note: You can set a seed for random value generator to obtain same output in
#'  different runs. However, the sample microdata included in the package will
#'  be different than this one, as it was generated from the corresponding
#'  python package for consistency in test output.
#'
#' @import data.table
#'
#' @param size Number of rows in the sample microdata. Default is 1000.
#' @param rkey_range The max range for record keys. Default is 255.
#' @param seed A seed for the random number generator
#'
#' @return A data.table containing randomly generated microdata and record keys
#' @export
#'
#' @examples
#' if (requireNamespace("data.table", quietly = TRUE)) {
#'   data.table::setDTthreads(1)
#' }
#' data <- generate_test_data(size = 1000)
#' data <- generate_test_data(size = 1000, rkey_range = 255, seed = 111)
generate_test_data <- function(size = 1000, rkey_range = 255, seed = NULL) {
  if(!is.null(seed)){
    set.seed(seed)
  }

  record_key_sample <- sample(0:rkey_range, size, replace = TRUE)

  var1 <- sample(
    x = 1:5,
    size = size,
    replace = TRUE,
    prob = c(0.25, 0.35, 0.20, 0.10, 0.10)
  )

  var2 <- sample(
    x = 1:2,
    size = size,
    replace = TRUE,
    prob = c(0.5, 0.5)
  )

  var3 <- sample(
    x = 1:4,
    size = size,
    replace = TRUE,
    prob = c(0.25, 0.35, 0.20, 0.20)
  )

  var4 <- sample(
    x = 1:4,
    size = size,
    replace = TRUE,
    prob = c(0.25, 0.35, 0.20, 0.20)
  )

  var5 <- sample(
    x = 1:10,
    size = size,
    replace = TRUE,
    prob = c(0.20, 0.15, 0.08, 0.15, 0.02, 0.025, 0.075, 0.10, 0.10, 0.10)
  )

  var6 <- sample(
    x = 1:5,
    size = size,
    replace = TRUE,
    prob = c(0.25, 0.35, 0.20, 0.10, 0.10)
  )

  var7 <- sample(
    x = 1:5,
    size = size,
    replace = TRUE,
    prob = c(0.25, 0.35, 0.20, 0.10, 0.10)
  )

  categories_ABCD <- c("A", "B", "C", "D")
  var8 <- sample(categories_ABCD, size = size, replace = TRUE)

  categories_ABCDEFGH <- c("A", "B", "C", "D", "E", "F", "G", "H")
  var9 <- sample(categories_ABCDEFGH, size = size, replace = TRUE)

  var10 <- sample(1:49, size, replace = TRUE)

  DT <- data.table(
    record_key = record_key_sample,
    var1 = var1,
    var2 = var2,
    var3 = var3,
    var4 = var4,
    var5 = var5,
    var6 = var6,
    var7 = var7,
    var8 = var8,
    var9 = var9,
    var10 = var10
  )

  return(DT)
}

---
title: "Cell Key Perturbation in R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Cell Key Perturbation in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(cellkeyperturbation)
```

## Summary

This method creates a frequency table which has had cell key perturbation 
applied to the counts to protect against disclosure. 

Cell key perturbation adds small amounts of noise to frequency tables. 
Noise is added to change the counts that appear 
in the frequency table by small amounts, for example a 14 is changed to a 15. 
This noise introduces uncertainty in the counts and makes it harder to identify
individuals, especially when taking the 'difference' between two similar
tables. It protects against the risk of disclosure by differencing since it 
cannot be determined whether a difference between two similar tables represents 
a real person, or is caused by the perturbation.

Cell Key Perturbation is consistent and repeatable, so the same cells are 
always perturbed in the same way.

It is expected that users will tabulate 1 to 4 variables for a particular 
geography level - for example, tabulate age by sex at local authority level. 

Full details of the methodology and statistical process flow are given in the [Methodology](#methodology) section.

Cell Key Perturbation method is available in [Python](https://github.com/ONSdigital/cell-key-perturbation) and 
[R](https://github.com/ONSdigital/cell-key-perturbation-R), each with integrated BigQuery functionality.

### BigQuery

The BigQuery version allows users to perform perturbation without reading raw data into local memory. 
The package creates the frequency table and runs perturbation with an SQL query. 
Then, it converts the final perturbed table into a `pandas.DataFrame`/`data.table` as an output. 

This will allow users to run the method on large datasets without breaking the memory limits. 

### Terminology

- ***Microdata*** - Data at the level of individual respondents
- ***Record key*** - A random number assigned to each record 
- ***Cell value*** - The number of records or frequency for a cell
- ***Cell key*** - The sum of record keys for a given cell
- ***pvalue*** - Perturbation value. The value of noise added to cells, e.g. +1, -1
- ***pcv*** - Perturbation cell value. This is an amended cell value needed to merge on the ptable
- ***ptable*** - Perturbation table. The look-up file containing the pvalues, this determines which cells get perturbed and by how much.

## Requirements 

- This method requires microdata and a perturbation table (ptable) file. 
- The microdata and the ptable both need to be supplied as `pandas.DataFrame`/`data.table` or BigQuery tables.
- The microdata must include a record key variable for cell key perturbation to be applied, unless `ons_id` will be used to create record key.
- You must use the provided ptable that corresponds to your microdata, e.g. `ptable_census21` for census 2021.

### Microdata and Record Keys

**Microdata** must be row-level, i.e. one row per statistical unit such as person or household. **Microdata** must contain one column per variable, which are expected to be categorical (they can be numeric but categorical is more suitable for frequency tables). 

**Record keys** should already be attached to the **microdata** as a column of integers in the range 0-255 or 0-4095, except certain ONS datasets with `ons_id`. 
The name of the **record key** column could change in different **microdata** tables. 
For example, **record key** columns in census data tables are named as `resident_record_key`, `household_record_key`, or `family_record_key` depending on the table type.

Certain ONS datasets contain `ons_id` column and use it as the basis for record keys to keep the perturbation consistent. 
If `ons_id` is available as a column in **microdata**, then **record keys** will be derived from `ons_id` by default. 
(This can be switched off by setting `use_existing_ons_id = False`)

The range of **record keys** should match the range of **cell keys** in the **ptable**. A warning message will be generated if those ranges do not match.

Cell Key Perturbation is consistent and repeatable, so the same cells are always perturbed in the same way. 

**The record keys need to be unchanged, changing the record keys would create inconsistent results and provide much less protection. You should use record keys attached to your microdata if provided instead of creating new ones to obtain consistent perturbation across different runs.**

### Perturbation Table (P-table)

The **perturbation table** contains the parameters which determine which cells are perturbed by how much and which are not (most cells are perturbed by +0). The **ptable** contains each possible combination of **cell key** (`ckey`) and **cell value** (`pcv`), and the **perturbation value** (`pvalue`) for each combination. 

A sample **ptable** that applies the '10-5 rule' is provided with the package and works with **record keys** in the range 0-255. This **ptable** will remove all cells below the threshold of 10, and round all others to the nearest 5. This provides more protection and will ensure safe outputs.

Other **ptables** may be available depending on the **microdata** used, for example census 2021 data will require the `ptable_census21` to be used and is based on cell keys in the range 0-255.

**You must use the specific ptable provided with the microdata you are working with to ensure sufficient and consistent protection, e.g. `ptable_census21` for census 2021.**


# User Instructions

## Installing the SML method

This method requires R version 3.5 or higher and uses the `data.table` package.

You can install the released version of `cellkeyperturbation` from CRAN:

```r
install.packages("cellkeyperturbation")
```

In your code you can load the cell key perturbation package using:

```r
library(cellkeyperturbation)
```

## Using the SML method

You can call the main functions for cell key perturbation with the following parameters:
```R
# for data.table
create_perturbed_table(data, ptable, geog, tab_vars, record_key, use_existing_ons_id, threshold)

# for BigQuery
create_perturbed_table_bigquery(con, data, ptable, geog, tab_vars, record_key, use_existing_ons_id, threshold)
```

Parameters specific for BigQuery version:

- **`con`** - (DBIConnection) - An active BigQuery connection created with `DBI::dbConnect()`
- **`data`** - (Microdata) - a `character` for the full name of micro-level `data` in BigQuery in "\<PROJECT>.\<DATASET>.\<TABLE>" format.
- **`ptable`** - (Perturbation table) - a `character` for the full name of `ptable` in BigQuery in "\<PROJECT>.\<DATASET>.\<TABLE>" format.

Parameters specific for data.table version:

- **`data`** - (Microdata) - a `data.table` containing the micro-level `data` to be tabulated and perturbed.
- **`ptable`** - (Perturbation table) - a `data.table` containing the `ptable` file which determines when 
perturbation is applied.

Common parameters for both versions:

- **`geog`** - (Geography) - a character vector giving the column name in `data` that contains the desired geography level you wish to tabulate at, e.g. `c("Local_Authority", "Ward")`. This can be the empty vector, `geog = c()`, if no geography level is required.
- **`tab_vars`** - (Variables to tabulate) - a character vector giving the column names in `data` of the variables to be tabulated e.g. `c("Age","Health","Occupation")`. This can also be the empty vector, `tab_vars = c()`. However, at least one of `tab_vars` or `geog` must be populated. If both are left blank an error message will be returned.
- **`record_key`** - a character containing the column name in `data` giving the record keys required for perturbation. If `ons_id` is available as a column in `data` and `use_existing_ons_id = TRUE`, set `record_key = NULL`, as record keys will be generated from `ons_id`.
- **`use_existing_ons_id`** - `TRUE` or `FALSE`, with a default of `TRUE`. If `ons_id` is available as a column in `data`, then record keys will be derived from `ons_id` by default.
- **`threshold`** - the value below which a count is suppressed (default 10).


## How to Use the Method in BigQuery

1. Create a BigQuery connection:
```R
# Import libraries
library(DBI)
library(bigrquery)

# Assign the project_id
project_id <- system("gcloud config get project", intern = TRUE)

# Create the connection using DBI
con <- DBI::dbConnect(
  bigrquery::bigquery(),
  project = project_id,
  bigint = "integer64"
)
```

2. Import `cellkeyperturbation` package and call the main function with parameters:
```R
library(cellkeyperturbation)

perturbed_table <- create_perturbed_table_bigquery(
  con        = con,
  data       = "<PROJECT>.<DATASET>.<microdata>",
  ptable     = "<PROJECT>.<DATASET>.<ptable>",
  geog       = c("Region"),
  tab_vars   = c("AgeGroup", "HealthStatus", "Occupation"),
  record_key = "Record_Key",
  threshold  = 10
)

# or
perturbed_table <- create_perturbed_table_bigquery(
  con        = con,
  data       = "<PROJECT>.<DATASET>.<microdata>",
  ptable     = "<PROJECT>.<DATASET>.<ptable>",
  geog       = c("Region"),
  tab_vars   = c("AgeGroup", "HealthStatus", "Occupation"),
  record_key = NULL,
  use_existing_ons_id = True
  threshold  = 10
)
```

3. The returned `perturbed_table` is a `data.table`. You need to drop disclosive columns before exporting the output from the secure data environment. Please refer to the [Interpreting the Output](#interpreting-the-output) and [Saving the Output](#saving-the-output) sections below for more details.
```R
perturbed_table[, c("pre_sdc_count", "ckey", "pcv", "pvalue") := NULL]
```


## Worked Example with Synthetic Data (`data.table`)

This is an example showing how to create a perturbed table from synthetic test data provided in the package (`micro` and `ptable_10_5`). You can access and view these data tables after loading the package.

```R
library(cellkeyperturbation)
View(micro)
View(ptable_10_5)
```

You can also generate different sample data or generate random record keys for testing purposes for your own test data with the following code:
```R
data = generate_test_data(size = 1000, rkey_range = 255, seed = 123)
ptable = generate_ptable_10_5_rule(ckey_range = 255)

library(data.table)
data <- fread("input_microdata.csv")
data = generate_random_rkey(data, rkey_range = 255, seed = 123)
```

- **`micro`**: A sample `data.table` containing randomly generated microdata and record keys.

Example rows of a microdata table are shown below:

 | record_key  | var1  | var5  | var8  | 
 |       :---  | :---- | :---- | :---- | 
 |      84     | 2     | 9     | D     | 
 |     108     | 1     | 9     | C     | 
 |     212     | 1     | 1     | D     | 
 |     212     | 2     | 2     | A     | 
 |      86     | 2     | 4     | A     | 

- **`ptable_10_5`**: A sample perturbation table (`data.table`) that defines the cell key perturbation rules. This specific table applies the ’10 to 5 rule’, which means a suppression threshold of *10* and rounding to the nearest *5*. In other words, this ptable will remove all cells under *10*, and round all others to the nearest *5*.

Example rows of a ptable are shown below:  

 | pcv  | ckey  | pvalue |
 |:---  | :---- | :----  |
 |   1  |    0  |    -1  | 
 |   1  |    1  |    -1  | 
 |   1  |    2  |    -1  | 
 | ...  |  ...  |   ...  |    
 | 750  |  255  |     0  |

Use the following code to generate the perturbed table using the sample microdata and perturbation table provided:

```R
perturbed_table <- create_perturbed_table(
  data       = micro,
  ptable     = ptable_10_5,
  geog       = c("var1"),
  tab_vars   = c("var5","var8"),
  record_key = "record_key",
  threshold  = 10
)
```


## Interpreting the Output

The output from the code is a `data.table` containing a frequency table with 
the counts having been affected by perturbation, as specified in the ptable. 

For most ptables, the most obvious effect will be that all counts lower than the 
threshold of 10 will have been removed. Suppressing counts below the threshold is a 
condition that need to be met when exporting data from IDS (Integrated Data Service)
and many other secure environments such as SRS (Secure Research Service).

The perturbation code will treat categories for missing data in the same way as it treats other categories. 
If you would like to exclude missing data from your outputs, you will need to remove the missing data categories either before or after applying the perturbation.

The table will be in the following format:

  | var1  | var5  | var8  | pre_sdc_count | ckey  | pcv   | pvalue | count  |
  |  :--- | :---- | :---- |       :----   | :---- | :---- | :----  | :----  | 
  |  1    |   1   |   A   |         10    | 173   |  10   |    0   |   10   | 
  |  1    |   1   |   B   |         10    |  88   |  10   |    0   |   10   | 
  |  1    |   1   |   C   |          7    | 180   |   7   |   -7   |  nan   | 
  |  1    |   1   |   D   |         14    |  66   |  14   |    1   |   15   | 
  |  1    |   2   |   A   |         11    | 190   |  11   |   -1   |   10   | 
  | ...   | ...   | ...   |        ...    | ...   | ...   |  ...   |  ...   | 
  

The table contains the variables used to summarise the 
data (in this example `var1`, `var5` & `var8`), and five other columns:

- `ckey` is the sum of record keys for each combination of variables.
- `pcv` is the perturbation cell value, the pre-perturbation count modulo 750.
- `pre_sdc_count` is the pre-perturbation count.
- `pvalue` is the perturbation applied to the original count, most commonly 
it will be 0. This is obtained from the ptable using a join on `ckey` and `pcv`.
- `count` is the post-perturbation count, the values to be output. It will be
set to `NA` if the value is suppressed for being below the threshold.

The columns you are most likely interested in are the variables, which 
are the categories you've summarised by, plus the `count` column.

**WARNING! - The `ckey`, `pcv`, `pre_sdc_count` and `pvalue` columns should be dropped before the 
contingency table is published. Otherwise, the perturbation can be unpicked and the output will be disclosive.**


## Saving the Output

Before the table is ready to be published the disclosive columns must be dropped. 
These cannot be output as they would allow for the perturbation to be unpicked. 
This code assumes that you have not changed the default column names; please update it if you have.

```R
perturbed_table[, c("pre_sdc_count", "ckey", "pcv", "pvalue") := NULL]
```

To save this dataframe as a csv file, you can use `data.table` fast write function:
```R
fwrite(perturbed_table, "perturbed_table.csv")
```


## Appendix - Help Pages

The package includes further help pages like Introduction to Cell Key Perturbation vignette 
and documentation for each function. You can access these pages by selecting the 
`cellkeyperturbation` package name in the packages tab of RStudio or using: 

```r
help(package=cellkeyperturbation)
```


# Methodology

The user is required to supply **microdata** and to specify which columns in the
data they want to tabulate by. They must also supply a **ptable** which will 
determine which cells get perturbed and by how much.

The **microdata** needs to contain a column for **record key**. **Record keys** are 
random, uniformly distributed integers within the chosen range. Previously, 
**record keys** between 0-255 have been used (as for census-2021). The method has 
been extended to also handle **record keys** in the range 0-4095 for the purpose of 
processing administrative data. 

It is expected that users will tabulate 1-4 variables for a particular geography 
level e.g. tabulate age by sex at local authority level. 

The `create_perturbed_table()` function counts how many rows in the data
contain each combination of categories e.g. how many respondents are of
each age category in each local authority area. The sum of the **record
keys** for each record in each cell is also calculated. Modulo 256 or 4096
of the sum is taken so this **cell key** is within range. The table now has 
**perturbation cell values** (`pcv`) and **cell keys** (`ckey`).

The **ptable** is merged with the data, matching on `pcv` and `ckey`. The merge 
provides a `pvalue` for each cell. The **post perturbation count** (`count`) 
is the **pre-perturbation count** (`pre_sdc_count`), plus the **perturbation value**
(`pvalue`). After this step, the counts have had the required perturbation 
applied. The output is the frequency table with the **post-perturbation count** (`count`) 
column. The result is that counts have been deliberately changed based on the 
**ptable**, for the purpose of disclosure protection.

To limit the size of the **ptable**, only 750 rows are used, and rows
501-750 are used repeatedly for larger cell values. E.g. instead of
containing 100,001 rows, when the cell value is 100,001 the 501st row
is used. Rows 501-750 will be used for cell values of 501-750, as well
as 751-1000, 1001-1250, 1251-1500 and so on. To achieve this effect an
alternative cell value column (`pcv`) is calculated which will be between 0-750.
For cell values 0-750 the `pcv` will be the same as the cell value. For
cell values above 750, the values are transformed by -1, modulo 250,
+501. This achieves the looping effect so that cell values 751, 1001,
1251 and so on will have a `pcv` of 501.

After cell key perturbation is applied, a **threshold** is applied so that any 
counts below the **threshold** will be suppressed (set to missing). The user can 
specify the value for the **threshold**, but if they do not, the default value of 
10 will be applied. Setting the **threshold** to zero would mean no suppression is 
applied.

As well as specifying the level of perturbation, the **ptable** can also be used 
to apply rounding, and a **threshold** for small counts. The example **ptable** 
supplied with this method, `ptable_10_5`, applies the 10_5 rule (supressing 
values less than 10 and rounding others to the nearest 5) for **record keys** 
in the range 0-255.


# Additional Information

The ONS Statistical Methods Library (statisticalmethodslibrary.ons.gov.uk) contains:

-	Further information about the methods including a link to the GitHub 
repository which contains detailed API information as part of the method code.
-	Information about other methods available through the library.


## License

Unless stated otherwise, the SML codebase is released under the MIT License. 
This covers both the codebase and any sample code in the documentation.
The documentation is available under the terms of the Open Government 3.0 
license. 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_ptable.R
\name{generate_ptable_10_5_rule}
\alias{generate_ptable_10_5_rule}
\title{Generate ptable (10-5 rule)}
\usage{
generate_ptable_10_5_rule(max_pcv = 750, ckey_range = 255)
}
\arguments{
\item{max_pcv}{Max value for pcv. Default is 750.}

\item{ckey_range}{The max range for cell keys. Default is 255.}
}
\value{
A data.table assigning a pvalue to each ckey and pcv combination
}
\description{
\code{generate_ptable_10_5_rule()} generates a sample p-table based on 10-5 rule,
which means a suppression threshold of 10 and rounding to the nearest 5.
}
\examples{
if (requireNamespace("data.table", quietly = TRUE)) {
  data.table::setDTthreads(1)
}
ptable <- generate_ptable_10_5_rule()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_for_na.R
\name{check_for_na}
\alias{check_for_na}
\title{Check perturbed table for missingness in tabulation variables}
\usage{
check_for_na(DT, cols)
}
\arguments{
\item{DT}{-- \code{data.table}
Perturbed frequency table}

\item{cols}{-- \verb{character vector}
Tabulation variables}
}
\value{
Warning message if any tabulation variable contain missing values
}
\description{
Check perturbed table for missingness in tabulation variables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_perturbed_table.R
\name{create_perturbed_table}
\alias{create_perturbed_table}
\title{Create a frequency table with cell key perturbation applied}
\usage{
create_perturbed_table(
  data,
  ptable,
  geog,
  tab_vars,
  record_key,
  use_existing_ons_id = TRUE,
  threshold = 10
)
}
\arguments{
\item{data}{A \code{data.table} containing the data to be tabulated and perturbed

The data should contain one row per statistical unit (person, household,
business or other) and one column per variable (age, sex, health status)}

\item{ptable}{A \code{data.table} containing the \code{ptable} file which determines
when perturbation is applied.}

\item{geog}{A \verb{character vector} giving the column name in \code{data} that
contains the desired geography level for the frequency table. This can be an
empty vector, \code{c()}, if no geography level is required.}

\item{tab_vars}{A \verb{character vector} giving the column names in \code{data} of the
variables to be tabulated.
This can be an empty vector, \code{c()}, provided a geography level is supplied.}

\item{record_key}{A \code{character} containing the column name in \code{data} giving
the record keys required for perturbation.
If data contains "ons_id" and use_existing_ons_id = TRUE,
set (record_key = NULL), as record key will be generated from "ons_id".}

\item{use_existing_ons_id}{A \code{logical} on whether to create record keys
from ons_id, if ons_id exists in data.
It will be irrelevant if microdata does not contain ons_id.
Default is TRUE.}

\item{threshold}{An \code{integer} specifying the value below which counts are
suppressed, with a default value of 10.}
}
\value{
Returns a \code{data.table} giving a frequency table which has had
cell key perturbation applied according to the ptable supplied.
}
\description{
\code{create_perturbed_table()} creates a frequency table which has had
cell key perturbation applied to the counts.
A p-table file needs to be supplied which determines which cells are
perturbed.
The data needs to contain a 'record key' variable which along with the
ptable allows the process to be repeatable and consistent.
}
\examples{
if (requireNamespace("data.table", quietly = TRUE)) {
  data.table::setDTthreads(1)
}

geog <- "var1"
tab_vars <- c("var5","var8")
record_key <- "record_key"
perturbed_table <- create_perturbed_table(micro,
                                          ptable_10_5,
                                          geog,
                                          tab_vars,
                                          record_key)

# Alternatively
perturbed_table <- create_perturbed_table(data = micro,
                                          ptable = ptable_10_5,
                                          geog = c(),
                                          tab_vars = c("var1","var5","var8"),
                                          record_key = "record_key",
                                          threshold = 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_inputs_before_perturbation.R
\name{validate_inputs}
\alias{validate_inputs}
\title{Validate Inputs Before Perturbation}
\usage{
validate_inputs(data, ptable, geog, tab_vars, record_key, threshold)
}
\arguments{
\item{data}{A \code{data.table} containing the data to be tabulated and perturbed

The data should contain one row per statistical unit (person, household,
business or other) and one column per variable (age, sex, health status)}

\item{ptable}{A \code{data.table} containing the \code{ptable} file which determines
when perturbation is applied.}

\item{geog}{A \verb{character vector} giving the column name in \code{data} that
contains the desired geography level for the frequency table. This can be an
empty vector, \code{c()}, if no geography level is required.}

\item{tab_vars}{A \verb{character vector} giving the column names in \code{data} of the
variables to be tabulated.
This can be an empty vector, \code{c()}, provided a geography level is supplied.}

\item{record_key}{A \code{character} containing the column name in \code{data} giving
the record keys required for perturbation.
If data contains "ons_id" and use_existing_ons_id = TRUE,
set (record_key = NULL), as record key will be generated from "ons_id".}

\item{threshold}{An \code{integer} specifying the value below which counts are
suppressed, with a default value of 10.}
}
\value{
Invisibly returns TRUE on success. Throws stop or Warning messages
if any validation fails.
}
\description{
Validates inputs for a perturbation process.
\itemize{
\item Validate type of input data & ptable
\item Validate other input arguments
\itemize{
\item Check that at least one variable specified for geog or tab_vars
\item Check geog and tab_vars are either character vectors or NULL
\item Check specified record_key is character vector or NULL
\item Check threshold is an integer and non-negative
}
\item Validate microdata and ptable contain required columns
\itemize{
\item Check data contain the specified geog, tab_vars & record_key
\item Check ptable contains required columns
}
\item Validate the range of record keys and cell keys
\item Validate data has sufficient records with record keys to apply perturbation
}
}
\examples{
if (requireNamespace("data.table", quietly = TRUE)) {
  data.table::setDTthreads(1)
}
validate_inputs(data = micro,
                ptable = ptable_10_5,
                record_key = "record_key",
                geog = c("var1"),
                tab_vars = c("var5","var8"),
                threshold = 10)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_perturbed_table_bigquery.R
\name{create_perturbed_table_bigquery}
\alias{create_perturbed_table_bigquery}
\title{Create a perturbed frequency table in BigQuery and return it as a data frame}
\usage{
create_perturbed_table_bigquery(
  con,
  data,
  ptable,
  geog,
  tab_vars,
  record_key,
  use_existing_ons_id = TRUE,
  threshold = 10,
  return_query = FALSE
)
}
\arguments{
\item{con}{--\code{DBIConnection}.
An active BigQuery connection created with \code{DBI::dbConnect()}}

\item{data}{--\code{character}.
BigQuery table name for microdata in full format:
\code{"<PROJECT>.<DATASET>.<TABLE>"}.
One row per statistical unit (person, household, business, etc.),
and one column per variable (e.g. age, sex, health status)}

\item{ptable}{--\code{character}.
BigQuery table name for the p-table in full format:
\code{"<PROJECT>.<DATASET>.<TABLE>"}.}

\item{geog}{--\verb{character vector}.
Column name containing the desired geography level for the frequency table.
e.g., \code{c("Region")} or \code{c("LocalAuthority")}.
Use \code{c()} if no geography breakdown required.}

\item{tab_vars}{--\verb{character vector}.
Column names to tabulate, e.g., \code{c("Age", "Health", "Occupation")}.}

\item{record_key}{--\code{character}.
Column name with record keys required for perturbation,
e.g., \code{"Record_Key"}.
If data contains "ons_id" and use_existing_ons_id = TRUE,
set (record_key = NULL), as record key will be generated from "ons_id".}

\item{use_existing_ons_id}{-- \code{logical}
Whether to create record keys from ons_id, if ons_id exists in data.
It will be irrelevant if microdata does not contain ons_id.
Default is TRUE.}

\item{threshold}{--\code{integer}.
Suppression threshold; perturbed counts below this value are suppressed.
Default 10.}

\item{return_query}{--\code{logical}.
If \code{TRUE}, returns the generated SQL query without executing it.
Default \code{FALSE}.}
}
\value{
\itemize{
\item When \code{return_query = FALSE}: a \code{data.table} containing the perturbed
frequency table, sorted by \code{geog} and \code{tab_vars}.
\item When \code{return_query = TRUE}: a character string containing the query.
}
}
\description{
This function runs the perturbation method fully in BigQuery (via SQL) and
only downloads the result, which allows handling large datasets efficiently.
}
\details{
Function workflow:
\enumerate{
\item Generate BigQuery SQL query to run perturbation on BigQuery
\item If return_query = TRUE, return the query text and exit:
otherwise, execute the rest
\item Validate inputs using BigQuery
\item Run perturbation using BigQuery
\item Convert perturbed table to data.table and sort
}

The query build by this function does the following when executed:
\itemize{
\item Computes counts and cell keys for each unique combination of geographic
and tabulation variables.
\item Includes zero-count cells by generating the full cartesian product of
variable combinations.
\item Calculates pcv by ensuring the rows of ptable 501-750 are reused for
cell values above 750.
\item Applies perturbation values from a perturbation table based on cell keys
and pseudo cell values (pcv).
\item Suppresses cells below a specified threshold by setting their perturbed
count to NULL.
}
}
\examples{
# --- Return query text without executing it ---
query <- create_perturbed_table_bigquery(
  con        = NULL,
  data       = "my-gcp-project.survey.microdata",
  ptable     = "my-gcp-project.sdc.ptable",
  geog       = c("Region"),
  tab_vars   = c("AgeGroup", "HealthStatus", "Occupation"),
  record_key = "Record_Key",
  threshold  = 10,
  return_query = TRUE
)
cat(query)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_record_key.R
\name{generate_record_key_from_ons_id}
\alias{generate_record_key_from_ons_id}
\title{Generate Record Key from ONS ID}
\usage{
generate_record_key_from_ons_id(data, record_key_col)
}
\arguments{
\item{data}{A \code{data.table} containing the \code{ons_id} column.}

\item{record_key_col}{A character string specifying the name of the new
record key column to create.}
}
\value{
A \code{data.table} with the new record key column added.
}
\description{
This function creates a new record key column by taking the modulo 4096 of
the \code{ons_id} column. It converts \code{ons_id} to numeric, preserving \code{NA} for
non-numeric values, and assigns the result as an integer.
}
\details{
\itemize{
\item The function checks that \code{data} is a \code{data.table}.
\item Non-numeric values in \code{ons_id} are converted to \code{NA}.
\item The record key is computed as \code{ons_id \%\% 4096} and stored as integer.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_record_key.R
\name{generate_random_rkey}
\alias{generate_random_rkey}
\title{Generate and attach random record keys to microdata}
\usage{
generate_random_rkey(data, rkey_range = 255, seed = NULL)
}
\arguments{
\item{data}{A data.table or data.frame containing the microdata}

\item{rkey_range}{The max range for record keys. Default is 255.}

\item{seed}{A seed for the random number generator}
}
\value{
A data.table with a new integer column \code{record_key}
}
\description{
\code{generate_random_key()} attaches randomly generated record keys to microdata
tables for testing purposes.
}
\examples{
library(data.table)
data <- data.table(id = 1:1000)
data <- generate_random_rkey(data, rkey_range = 255, seed = 2005)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ptable_10_5}
\alias{ptable_10_5}
\title{Perturbation table}
\format{
A data.table containing 192000 observations of 3 variables
}
\usage{
data(ptable_10_5)
}
\description{
A data set containing the rules to apply cell key perturbation with a
threshold of 10, and rounding to base 5. In other words, counts less than
10 will be removed, and all others will be rounded to the nearest 5.
}
\details{
\itemize{
\item pcv. perturbation cell value (1-750)
\item ckey. cell key value (0-255)
\item pvalue. perturbation value to be applied
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_data.R
\name{generate_test_data}
\alias{generate_test_data}
\title{Generate sample microdata}
\usage{
generate_test_data(size = 1000, rkey_range = 255, seed = NULL)
}
\arguments{
\item{size}{Number of rows in the sample microdata. Default is 1000.}

\item{rkey_range}{The max range for record keys. Default is 255.}

\item{seed}{A seed for the random number generator}
}
\value{
A data.table containing randomly generated microdata and record keys
}
\description{
\code{generate_test_data()} creates a sample microdata containing randomly
generated microdata columns and record keys for testing purposes.

Note: You can set a seed for random value generator to obtain same output in
different runs. However, the sample microdata included in the package will
be different than this one, as it was generated from the corresponding
python package for consistency in test output.
}
\examples{
if (requireNamespace("data.table", quietly = TRUE)) {
  data.table::setDTthreads(1)
}
data <- generate_test_data(size = 1000)
data <- generate_test_data(size = 1000, rkey_range = 255, seed = 111)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{micro}
\alias{micro}
\title{Example data (micro)}
\format{
A data.table containing 1000 observations of 11 variables
}
\usage{
data(micro)
}
\description{
A data set containing randomly generated data to showcase the cell key
perturbation method.
}
\details{
\itemize{
\item record_key. record key value (0-255)
\item var1. example variable 1 (1-5)
\item var2. example variable 2 (1,2)
\item var3. example variable 3 (1-4)
\item var4. example variable 4 (1-4)
\item var5. example variable 5 (1-10)
\item var6. example variable 6 (1-5)
\item var7. example variable 7 (1-5)
\item var8. example variable 8 (A-D)
\item var9. example variable 9 (A-H)
\item var10. example variable 10 (1-49)
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_inputs_before_perturbation.R
\name{validate_inputs_bigquery}
\alias{validate_inputs_bigquery}
\title{Validate Inputs Before Perturbation using BigQuery}
\usage{
validate_inputs_bigquery(
  con,
  data,
  ptable,
  geog,
  tab_vars,
  record_key,
  use_existing_ons_id,
  threshold
)
}
\arguments{
\item{con}{--\code{DBIConnection}.
An active BigQuery connection created with \code{DBI::dbConnect()}}

\item{data}{--\code{character}.
BigQuery table name for microdata in full format:
\code{"<PROJECT>.<DATASET>.<TABLE>"}.
One row per statistical unit (person, household, business, etc.),
and one column per variable (e.g. age, sex, health status)}

\item{ptable}{--\code{character}.
BigQuery table name for the p-table in full format:
\code{"<PROJECT>.<DATASET>.<TABLE>"}.}

\item{geog}{--\verb{character vector}.
Column name containing the desired geography level for the frequency table.
e.g., \code{c("Region")} or \code{c("LocalAuthority")}.
Use \code{c()} if no geography breakdown required.}

\item{tab_vars}{--\verb{character vector}.
Column names to tabulate, e.g., \code{c("Age", "Health", "Occupation")}.}

\item{record_key}{--\code{character}.
Column name with record keys required for perturbation,
e.g., \code{"Record_Key"}.
If data contains "ons_id" and use_existing_ons_id = TRUE,
set (record_key = NULL), as record key will be generated from "ons_id".}

\item{use_existing_ons_id}{-- \code{logical}
Whether to create record keys from ons_id, if ons_id exists in data.
It will be irrelevant if microdata does not contain ons_id.
Default is TRUE.}

\item{threshold}{--\code{integer}.
Suppression threshold; perturbed counts below this value are suppressed.
Default 10.}
}
\value{
Invisibly returns TRUE on success. Throws stop or Warning messages
if any validation fails.
}
\description{
Validates BigQuery inputs for a perturbation process.
\itemize{
\item Validate input arguments
\itemize{
\item Check that at least one variable specified for geog or tab_vars
\item Check geog and tab_vars are either character vectors or NULL
\item Check specified record_key is character vector or NULL
\item Check threshold is an integer and non-negative
}
\item Validate microdata and ptable contain required columns
\itemize{
\item Check data contain the specified geog, tab_vars & record_key
\item Check ptable contains required columns
}
\item Validate the range of record keys and cell keys
\item Validate data has sufficient records with record keys to apply perturbation
}
}
